#!/dev/null
# vim: sts=4 et ts=4 encoding=utf-8

__module_name__ = "hilight-ignore++"
__module_version__ = "0.1"
__module_description__ = "Advanced highlighting and ignoring options"

import re
import os
import os.path
import json
import errno
from time import time
from datetime import datetime

import xchat

confpath = os.path.join(xchat.get_info('xchatdir'), 'scripts_conf', 'epsy')

catch_list = {
    'change nick': ('nn', None),
    'channel action': ('nt', 'Channel Action Hilight'),
    'channel ban': ('n', None),
    'channel dehalfop': ('nn', None),
    'channel deop': ('nn', None),
    'channel devoice': ('nn', None),
    'channel exempt': ('n', None),
    'channel half-operator': ('nn', None),
    'channel invite': ('n ', None),
    'channel message': ('nt', 'Channel Msg Hilight'),
    'channel mode generic': ('n', None),
    'channel notice': ('n', None),
    'channel operator': ('nn', None),
    'channel remove exempt': ('n', None),
    'channel remove invite': ('n', None),
    'channel remove keyword': ('n', None),
    'channel remove limit': ('n', None),
    'channel set key': ('n', None),
    'channel set limit': ('n', None),
    'channel unban': ('n', None),
    'channel voice': ('nn', None),
    'ctcp generic': ('tn', None),
    'ctcp generic to channel': ('tn', None),
    'ctcp sound': ('tn', None),
    'ctcp sound to channel': ('tn', None),
    'dcc chat offer': ('n', None),
    'dcc generic offer': ('tn', None),
    'dcc send offer': ('tnt', None),
    'invited': ('cns', None),
    'join': ('ncm', None),
    'kick': ('nnct', None),
    'notice': ('nt', None),
    'part': ('nmc', None),
    'part with reason': ('nmct', None),
    'private action': ('nt', None),
    'private action to dialog': ('nt', None),
    'private message': ('nt', None),
    'private message to dialog': ('nt', None),
    'quit': ('ntm', None),
    'raw modes': ('nt', None),
    'topic': ('ct', None),
    'topic change': ('ntc', None),
    }

class ListOfEverything(list):
    def __contains__(self, other):
        return True

def pmsg(msg):
    xchat.prnt('\00309hi++\003\t' + msg)

PATTERN_PARTIAL=0
PATTERN_FULL=1
PATTERN_REGEXP=2

def text_to_re(text, text_type=PATTERN_PARTIAL):
    if not text:
        return None

    if text_type == PATTERN_REGEXP:
        ret = re.compile(text)
    elif text_type == PATTERN_FULL:
        ret = re.compile('\W*{0}\W*'.format(re.escape(text.lower())))
    elif text_type == PATTERN_PARTIAL:
        ret = re.compile('{0}'.format(re.escape(text.lower())))

    return re.compile(ret)

def make_ircmask(text):
    if not text:
        return None
    if '*' in text or '?' in text or '!' in text or '@' in text:
        return text
    else:
        return text + '!*@*'

def ircmask_to_re(text):
    if not text:
        return None

    return re.compile(
        re.escape(text).replace('\\*', '.*').replace('\\?', '.?')
        )

def get_mask(nick):
    nick = nick.lower()
    if xchat.get_info('channel')[0] == '#':
        for user in xchat.get_list('users'):
            if nick == user.nick.lower():
                return nick + '!' + user.host
        return nick + '!XCHATUNKNOWN@XCHATUNKNOWN'
    else:
        return nick + '!' + xchat.get_info('topic')

reply = re.compile(r'(\S+?)\W\s')

class MessagePattern(object):
    def __init__(self, mask, text, text_type, msg_types, networks,
            channels, exemption, replies, expire):
        self.mask = make_ircmask(mask)
        self.mask_re = ircmask_to_re(self.mask)
        self.re = text_to_re(text, text_type)
        self.text = text
        self.text_type = text_type
        self.msg_types = msg_types
        self.networks = networks
        self.channels = channels
        self.exemption = exemption
        self.replies = replies
        self.expire = expire

    def to_json(self):
        ret = {
            'mask': self.mask,
            'text': self.text,
            'text_type': self.text_type,
            'exemption': self.exemption,
            'replies': self.replies,
            'expire': self.expire
            }
        for name in ('msg_types', 'networks', 'channels'):
            val = getattr(self, name)
            if type(val) == ListOfEverything:
                val = []
            ret[name] = val
        return ret

    @classmethod
    def from_json(cls, obj):
        for name in ('msg_types', 'networks', 'channels'):
            if not obj[name]:
                obj[name] = ListOfEverything()
        return cls(
            obj['mask'], obj['text'], obj['text_type'],
            obj['msg_types'], obj['networks'], obj['channels'],
            obj['exemption'], obj['replies'], obj['expire']
            )

    def expired(self):
        if self.expire and self.expire < time():
            return True
        else:
            return False

    def matches(self, words, level):
        if self.expired():
            return False

        if self.applies_here(level):
            for i, t in enumerate(catch_list[level][0]):
                if t == 'n' and self.match_nick(words[i]) \
                or t == 't' and self.match_text(words[i]):
                    return True

        return False

    def applies_here(self, level):
        if level not in self.msg_types:
            return False

        if type(self.networks) == ListOfEverything \
        and type(self.channels) == ListOfEverything:
            return True

        ctx = xchat.get_context()

        if ctx.get_info('channel') not in self.channels:
            return False

        if ctx.get_info('network') not in self.networks:
            return False

        return True

    def match_level(self, level):
        return level in self.msg_types

    def match_nick(self, nick):
        if self.mask_re:
            mask = get_mask(nick)
            return mask and self.mask_re.match(mask)
        else:
            return False

    def match_text(self, text):
        text = text.lower()
        return (
                self.re and self.re.match(text)
            ) or (
                self.replies and self.match_reply(text)
                )

    def match_reply(self, text):
        try:
            nick = reply.match(text).group(1)
        except AttributeError:
            return False

        return self.match_nick(nick)

class ListManager(object):
    typename = NotImplemented
    filename = NotImplemented
    command_add = NotImplemented
    command_del = NotImplemented

    def item_str(item):
        return NotImplemented

    def list(self, n=0):
        if self.items:
            if n > 0:
                l = ((n-1, self.items[n-1]),)
            else:
                pmsg("{0}s:".format(self.typename.capitalize()))
                l = enumerate(self.items)
            for i, x in l:
                xchat.prnt('{0: >3} {1}'.format(i+1, self.item_str(x)))
        elif not n:
            pmsg("No {0}s to list yet. See \x02/HELP {1}\x02 "
                "for info on setting {0}s.".format(self.typename, self.command_add))
        else:
            pmsg("No {0}s left.".format(self.typename))

    def build_item(self, rest, *args, **kwargs):
        return NotImplemented

    def handle_switch(self, switch, next_word, opts):
        return NotImplemented

    def add_cmd(self, words, word_eol, userdata):
        skip_next = False
        opts = {}
        for i, word in enumerate(words[1:]):
            if not skip_next:
                if word[0] == '-':
                    try:
                        switch = word[1:]
                        skip_next = self.handle_switch(switch, words[i+2], opts)
                    except IndexError:
                        pmsg('Missing argument to -{0}.'.format(switch))
                        return False
                    except ValueError:
                        return False
                else:
                    rest = words[i+1:]
                    break
            else:
                skip_next = False

        try:
            item = self.build_item(rest, **opts)
            if not item:
                return False
            self.items.append(item)
            return len(self.items)
        except re.error as e:
            pmsg('Error while reading regular expression: {0}'.format(e))
            return False

    def del_cmd(self, word, word_eol, userdata):
        try:
            n = int(word[1])
        except ValueError:
            pmsg('Number required.')
            return False

        if n <= len(self.items):
            del self.items[n - 1]
        else:
            pmsg('There are only "{0}" {1} rules.'.format(
                len(self.items), self.typename))
        return -1

    def cmd(self, word, word_eol, userdata):
        try:
            if len(word) <= 1: # No arguments, show a list.
                self.list()
            else:
                if word[0].upper() == self.command_del:
                    ret = self.del_cmd(word, word_eol, userdata)
                else:
                    ret = self.add_cmd(word, word_eol, userdata)

                if ret:
                    self.list(ret)

        finally:
            return xchat.EAT_ALL

    def matches(self, words, level):
        matched = False
        to_remove = []
        for pattern in self.items:
            if pattern.expired():
                to_remove.append(pattern)
                continue
            if pattern.exemption:
                if pattern.matches(words, level):
                    return False
            elif not matched:
                matched = pattern.matches(words, level)
        for pattern in to_remove:
            self.items.remove(pattern)
        return matched

    def caught(self, words, level):
        return NotImplemented

    def catch(self, words, word_eol, level):
        if self.matches(words, level):
            return self.caught(words, level)

    def save(self, *args):
        try:
            os.makedirs(confpath)
        except os.error as e:
            if e.errno != errno.EEXIST:
                pmsg('Error while creating config directory: {0}'.format(e))
                return xchat.EAT_XCHAT

        try:
            fp = file(os.path.join(confpath, self.filename+'.json'), 'w')
        except IOError as e:
            pmsg('Error while opening config for write: {0}'.format(e))
            return xchat.EAT_XCHAT

        json.dump(self.items, fp,
            default=MessagePattern.to_json,
            indent=4,
            separators=(',', ': ')
            )

        if len(args) > 1:
            pmsg('Saved {0}s to {1}'.format(
                self.typename, os.path.join(confpath, self.filename+'.json')))

        return xchat.EAT_XCHAT

    def load(self, *args):
        try:
            fp = file(os.path.join(confpath, self.filename+'.json'))
        except IOError as e:
            if e.errno != errno.ENOENT: # Don't warn on file not found
                pmsg('Error while opening config for read: {0}'.format(e))
        else:
            try:
                self.items = json.load(fp, object_hook=MessagePattern.from_json)
            except:
                pass

        if args:
            pmsg('Loaded {0}s from {1}'.format(
                self.typename, os.path.join(confpath, self.filename+'.json')))

        return xchat.EAT_XCHAT

    def __init__(self):
        self.items = []
        xchat.hook_command(self.command_add, self.cmd, help=self.__doc__)
        xchat.hook_command(self.command_del, self.cmd,
        help="""Usage: /{0} <n>

        Deletes the <n>th {1} rule. Type /{2} for a list.
    """.format(self.command_del, self.typename, self.command_add)
        )

        xchat.hook_command('SAVE', self.save)
        xchat.hook_command('RELOAD', self.load)
        xchat.hook_unload(self.save)

        for msg_type in catch_list:
            xchat.hook_print(msg_type, self.catch, msg_type)

class HilightManager(ListManager):
    """Usage: HILIGHT [-except] [-full | -regexp] [-mask <mask>] [-levels <levels>] [-networks <networks>] [-channels <channels>] [-time <seconds>] <text>

    Sets <text> in messages to cause highlighting.

    Flags:

        -except: When this rule is matched, *never* hilight the message.

        -full: <text> must match full words.
        -regexp: <text> is a regular expression. (See http://docs.python.org/library/re.html)

        -mask: Match only messages from <mask>.

        -levels: Comma-separated list of text events that can trigger the hilight. (Default: "channel message, channel action, private message, private action, private message to dialog, private action to dialog, notice")
        -networks: Comma-separated list of networks this rule should apply on.
        -channels: Comma-separated list of channels this rule should apply on.

        -time: Time after which the rule is removed.

    Examples:

    Highlight lines that have "mynick" word:
        /HILIGHT -full mynick

    Highlight all messages from "mynick":
        /HILIGHT -mask mynick!*@* -regexp .*

    Highlight lines that have "@admins" in channel #mychannel
        /HILIGHT -channels #mychannel @admins

    See also: /HELP DEHILIGHT, Settings -> Advanced -> Text Events...
"""

    typename = 'highlight'
    filename = 'hilight'
    command_add = 'HILIGHT'
    command_del = 'DEHILIGHT'

    default_levels = (
                'channel message', 'channel action',
                'private message', 'private action',
                'private message to dialog', 'private action to dialog',
                'notice'
                )

    def item_str(self, item):
        ret = []
        if item.exemption:
            ret.append('-except')
        if item.channels:
            ret.extend(item.channels)
        if item.networks:
            ret.extend(item.networks)
        if item.text_type != PATTERN_PARTIAL:
            if item.text_type == PATTERN_FULL:
                ret.append('-full')
            elif item.text_type == PATTERN_REGEXP:
                ret.append('-regexp')
        if item.mask and item.mask != '*':
            ret.append('-mask ' + item.mask)
        if tuple(item.msg_types) != self.default_levels:
            ret.append('-levels "{0}"'.format(','.join(item.msg_types)))
        if item.expire:
            ret.append('-expire ' + datetime.fromtimestamp(item.expire).isoformat())

        return item.text + str(
                (ret and ('  ' + ' '.join(ret)))
                or ''
            )

    def build_item(self,
            rest,
            mask=None,
            text_type=PATTERN_PARTIAL,
            levels=None,
            networks=ListOfEverything(),
            channels=ListOfEverything(),
            exemption=False,
            replies=False,
            expire=None
            ):
        if not rest:
            raise ValueError
        text = ' '.join(rest)
        if not levels:
            levels = self.default_levels
        return MessagePattern(mask, text, text_type, levels, networks,
                channels, exemption, replies, expire)

    def handle_switch(self, switch, next_word, opts):
        if switch == 'mask':
            opts['mask'] = next_word.strip()
            return True
        elif switch == 'regexp':
            opts['text_type'] = PATTERN_REGEXP
        elif switch == 'full':
            opts['text_type'] = PATTERN_FULL
        elif switch == 'levels':
            opts['levels'] = next_word.split(',')
            for i, level in enumerate(opts['levels']):
                level = level.lower().strip()
                opts['levels'][i] = level
                if level not in catch_list:
                    pmsg('Unknown level "{0}"'.format(level))
                    raise ValueError
            return True
        elif switch == 'networks':
            opts['networks'] = [w.strip() for w in next_word.split(',')]
            return True
        elif switch == 'channels':
            opts['channels'] = [w.strip() for w in next_word.split(',')]
            return True
        else:
            pmsg('Unknown flag "-{0}".'.format(switch))
            raise ValueError

    def caught(self, words, level):
        xchat.command('gui color 3')
        if xchat.get_prefs('input_tray_hilight'):
            xchat.command('gui flash')
        newlevel = catch_list[level][1]
        if newlevel:
            xchat.emit_print(newlevel, *words)
            return xchat.EAT_ALL
        else:
            return

class IgnoreManager(ListManager):
    """Usage: IGNORE [-except] [-replies] [[-full | -regexp] -pattern <pattern>] [-networks <networks>] [-channels <channels>] [-time <seconds>] <mask> [<levels>]

    Ignore messages of type(s) <levels> from <mask>.

    Flags:

        -except: When this rule is matched, *never* ignore the message.

        -replies: Also ignore replies to <mask>. Does not work with patterns.

        -full: <pattern> must be a full word.
        -regexp: <pattern> is a regular expression. (See http://docs.python.org/library/re.html)
        -pattern: Match only when the message text contains <pattern>

        -networks: Comma-separated list of networks this rule should apply on.
        -channels: Comma-separated list of channels this rule should apply on.
        -levels: Comma-separated list of text events that can trigger the hilight. (Default: "channel message, channel action, private message, private action, private message to dialog, private action to dialog, notice")

        -time: Time after which the rule is removed.

    Examples:

    Ignore people nicknamed "someone" and replies to them:
        /IGNORE -replies someone

    Ignore people from a defined host:
        /IGNORE *!*@*.aol.com

    Ignore away chatter:
        /IGNORE -regexp -pattern "is (away|gone|back)" * channel action
        /IGNORE *away* change nick
        /IGNORE *back* change nick
        /IGNORE *zzz* change nick
        /IGNORE *work* change nick

    See also: /HELP UNIGNORE, Settings -> Advanced -> Text Events...
"""
    typename = 'ignore'
    filename = 'ignore'
    command_add = 'IGNORE'
    command_del = 'UNIGNORE'

    def item_str(self, item):
        ret = []
        if item.exemption:
            ret.append('-except')
        if item.replies:
            ret.append('-replies')
        if item.text:
            if item.text_type == PATTERN_FULL:
                ret.append('-full')
            elif item.text_type == PATTERN_REGEXP:
                ret.append('-regexp')
            ret.append('-pattern ' + item.text)
        if item.networks:
            ret.append(' '.join(item.networks))
        if item.channels:
            ret.append(' '.join(item.channels))
        if item.expire:
            ret.append('-expire ' + datetime.fromtimestamp(item.expire).isoformat())

        return item.mask + ': ' + str(( type(item.msg_types) == ListOfEverything
                        and 'ALL'
            ) or ','.join(item.msg_types)) + str(
                (ret or '') and ('  ' + ' '.join(ret))
            )


    def build_item(self,
            rest,
            text=None,
            text_type=PATTERN_PARTIAL,
            levels=ListOfEverything(),
            networks=ListOfEverything(),
            channels=ListOfEverything(),
            exemption=False,
            replies=False,
            expire=None
            ):
        if not rest:
            pmsg("Need a channel or mask to ignore.")
            return False

        mask = rest.pop(0)

        if mask[0] == '#':
            channels = [w.strip() for w in mask.split(',')]
            mask = '*'

        rest = ' '.join(rest).strip()
        if rest:
            levels = [w.strip() for w in rest.split(',')]
            for i, level in enumerate(levels):
                level = level.lower().strip()
                levels[i] = level
                if level not in catch_list:
                    pmsg('Unknown level "{0}"'.format(level))
                    return False


        return MessagePattern(mask, text, text_type, levels, networks,
                channels, exemption, replies, expire)

    def handle_switch(self, switch, next_word, opts):
        if switch == 'except':
            opts['exemption'] = True
        elif switch == 'regexp':
            opts['text_type'] = PATTERN_REGEXP
        elif switch == 'full':
            opts['text_type'] = PATTERN_FULL
        elif switch == 'pattern':
            opts['text'] = next_word
            return True
        elif switch == 'networks':
            opts['networks'] = [w.strip() for w in next_word.split(',')]
            return True
        elif switch == 'channels':
            opts['channels'] = [w.strip() for w in next_word.split(',')]
            return True
        elif switch == 'replies':
            opts['replies'] = True
        elif switch == 'time':
            try:
                opts['expire'] = time() + int(next_word)
            except ValueError:
                pmsg('Expected a number after -time.')
                raise ValueError
            return True
        else:
            pmsg('Unknown flag "-{0}".'.format(switch))
            raise ValueError

    def caught(self, words, levels):
        return xchat.EAT_ALL

hilight = HilightManager()
ignore = IgnoreManager()

hilight.load()
ignore.load()

xchat.prnt("Loaded {0} {1}".format(__module_name__, __module_version__))
